/*  hdmp.c - free software (GPL License) by Jens Goedeke

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Release info:
    V0.1 / 10.10.2000   first release
    V0.2 / 29.12.2003   new options -r and -s
    V1.0 / 10.01.2006   new options -f, -t and -b
    V1.1 / 02.08.2006   filename can be given in any position of the option sequence

    new release can be found at http://www.goedeke.net/tools_eng.html

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/stat.h>

/* Typendeklaration */
#define DWORD unsigned long
#define WORD  unsigned int
#define BYTE  unsigned char

/* Zeichenpuffer (Auslese-String aus Datei) */
BYTE buffer[128];

/* Zeichen das anstelle eines Steuerzeichens im Ascii Dump ausgegeben */
/* werden soll */
BYTE REPLACE = '.';
char SEPARATOR[2];


/**  */
DWORD filesize( char * filename )
/**  */
/* Groesse einer Datei ermitteln */
/* Bei Erfolg >0 */
/* Bei Misserfolg =0 */
{
    struct stat statbuf;
    if ( stat( filename, & statbuf ) ) return 1;
    else
        return ( ( DWORD )statbuf.st_size );
}



/**  */
int printline( DWORD Position, BYTE Count, BYTE Max )
/**  */
/* Ausgabe von COUNT Zeichen als Monitorzeile ( Adresse/Hex-Werte/Dump ) */
/* POSITION wird als 32 Bit Hexwert ausgegeben (16Bit:16Bit) */
/* MAX ist die Anzahl der maximalen Zeichen. */
/* Ist COUNT=MAX erfolgt keine Anpassung */
/* Ist COUNT<MAX wird die Differenz in Leerzeichen in die Ausgabe- */
/* Zeile eingefuegt. ( Damit die letzte Zeile die gleiche Formatierung */
/* erhaelt wie die vorherigen ) */
{
    char line[1000];
    /* Puffer fuer die fertige Zeile */
    char dummy[1000];
    /* Puffer fuer temporaere Zeilenfragmente */
    int c;

    if ( Max < Count || Max == 0 || Count == 0 ) return 1;


    /* Adresse ausgeben */
    sprintf( line, "%04X%04X %s", ( WORD )( Position >> 16 ), ( WORD )( Position & 65535 ), SEPARATOR );

    /* Hexwerte ausgeben */
    for ( c = 0; c < Count; c++ ) {
        sprintf( dummy, "%02X%s", buffer[c], SEPARATOR );
        strcat( line, dummy );
    }

    /* ggf. Zeile mit Leerzeichen auffuellen */
    for ( c = Count; c < Max; c++ ) {
        strcat( line, "  " );
        strcat( line, SEPARATOR );
    }

    /* Ascii-Dump ausgeben (ohne Controlzeichen) */
    /* Controlzeichen werden als Zeichen REPLACE ausgegeben */
    for ( c = 0; c < Max; c++ ) {
        if ( c >= Count ) dummy[c] = ' ';
        else if ( !iscntrl( buffer[c] ) && ( SEPARATOR[0] == ' ' || buffer[c] != SEPARATOR[0] ) ) dummy[c] = buffer[c];
        else
            dummy[c] = REPLACE;
    }
    /* NULL-Zeichen einfuegen */
    dummy[Max] = '\0';

    /* Der String wird nicht mehr an die Zeile LINE angefuegt, sondern separat */
    /* ausgegeben. */

    /* Ausgabe auf STDOUT */
    printf( "%s %s %s\n", line, dummy, SEPARATOR );
    return 0;
}

/* Programm beenden */
void exitproc() {
    printf( "\n" );
    printf( "------------------------------------------------------------\n" );
    printf( "hdmp V1.1 - hex/ascii dump generator\n" );
    printf( "------------------------------------------------------------\n" );
    printf( "hdmp dumps a given file to standard-output in the format:\n" );
    printf( "hex-fileposition / n * hexvalue / n * ascii chars\n" );
    printf( "\n" );
    printf( "Usage :\n" );
    printf( "hdmp <sourcefile> [-n:num] [-s:sep] [-r:rep] [-f:from]  \n" );
    printf( "                  [-l:lines | -b:bytes | -t:to]\n" );
    printf( "num   is the number of values per dump line (default=16)\n" );
    printf( "      range of [num] is 8 to 64 step 2\n" );
    printf( "sep   is the separator between the values (default is <SPACE>) \n" );
    printf( "rep   is the character that is used to replace illegal chars (default is '.')\n" );
    printf( "from  is the first position in file which should be dumped\n" );
    printf( "lines is count of lines that should be dumped (-n option must be set first)\n" );
    printf( "bytes is count of bytes that should be dumped\n" );
    printf( "to    is the last position in file which should be dumped\n" );
    printf( "\n" );
    printf( "hdmp - open source by Jens Goedeke 2000,2003 & 2006\n" );
    printf( "new release can be found at http://www.goedeke.net/tools_eng.html\n\n");
    exit (1);
}


/**  */
int main( int argc, char * argv[] )
/**  */
/* Hauptprogramm */
{
    FILE * in;
    DWORD Size = 0, Position = 0, From = 0, To = 0, Bytes = 0, TempSize, Lines = 0;
    BYTE Max = 16, Count;
    int x, i, l;
    char dummy[256];
    char option;
    char * filename = NULL;

    SEPARATOR[0] = ' '; SEPARATOR[1] = '\0';
    if (argc<2) exitproc();

    /* Argumente testen : 1. Ist eine Ausgabewert-Anzahl Aenderung vorhanden 2. Wurde ein neuer Separator eingegeben
    3. Wurde ein neues Replacezeichen eingegeben */
    for ( i = 1; i < argc; i++ ) {
        // printf ("%s\n",argv[i]);
        l = strlen( argv[i] );
        if ( l > 2 && argv[i] [0] == '-' ) {
            /* Anzahl der Werte ermitteln */
            option = argv[i] [1];
            strcpy( dummy, & ( argv[i] [3] ) );
            if ( option == 'n' ) {
                x = atoi( dummy );
                if ( ( x % 2 ) == 0 && x <= 64 && x >= 8 ) Max = x;
                continue;
            }
            /* Neuen Separator festlegen */
            if ( option == 's' && dummy[0] ) {
                SEPARATOR[0] = dummy[0];
                continue;
            }

            /* Neues Ersetzungszeichen festlegen */
            if ( option == 'r' && dummy[0] ) {
                REPLACE = dummy[0];
                continue;
            }

            if ( option == 'b' && strlen( dummy ) ) {
                Bytes = atol( dummy );
                if ( !Bytes ) {
                    printf( "Error : Byte count must be greater or equal to 1\n" );
                    exit( 1 );
                }
                continue;
            }

            if ( option == 'f' && strlen( dummy ) ) {
                From = atol( dummy );
                if ( !From ) {
                    printf( "Error : From position must be greater or equal to 1\n" );
                    exit( 1 );
                }
                continue;
            }

            if ( option == 't' && strlen( dummy ) ) {
                To = atol( dummy );
                if ( !To ) {
                    printf( "Error : To position must be greater or equal to 1\n" );
                    exit( 1 );
                }
                continue;
            }

            if ( option == 'l' && strlen( dummy ) ) {
                Lines = atol( dummy );
                if ( !Lines ) {
                    printf( "Error : Line count must be greater or equal to 1\n" );
                    exit( 1 );
                }
                continue;
            }

            printf( "Unknown Option : %c / %s\n", option, dummy );
            exit( 2 );
        }
        else {
            filename = argv[i];
        }
    }
    if ( strlen( filename ) < 1 ) {
        printf( "Error : Filename not given \n" );
        exit( 3 );
    }

    if ( Bytes > 0 ) To = From + Bytes;
    /* Ermitteln ob die Quelldatei vorhanden ist (Size>0) */
    if ( argc > 1 ) Size = filesize( filename );
    if ( Lines > 0 && From + ( Max * Lines ) < Size ) To = From + ( Max * Lines );
    if ( From > Size ) {
        printf( "Error : From (%lu) is greater than file size (%lu)\n", From, Size );
        exit( 3 );
    }
    if ( From > To ) {
        printf( "Error : From (%lu) is greater than To (%lu)\n", From, To, Bytes );
        exit( 2 );
    }
    if ( To == 0 || To > Size ) To = Size;
    /* Quell-Datei ausgeben */
    if ( Size > 0 && ( in = fopen( filename, "rb" ) ) != NULL ) {
        /* printf( "%lu Bytes Hexdump of < %s > \n", ( DWORD )Size,filename ); */
        do {
            if ( From > 0 && Position < From && Position + Max > From ) {
                TempSize = From - Position;
            }
            else
                TempSize = Max;

            if ( To > 0 && Position < To && Position + TempSize > To ) {
                TempSize = To - Position;
            }
            Count = fread( buffer, 1, TempSize, in );
            if ( Position >= From ) {
                printline( Position, Count, Max );
            }
            Position += Count;
        }
        while ( Position < To && Count == TempSize );
        fclose( in );
        return 0;
    }
    else {
        printf( "Error: File does not exist or has zero length !\n" );
		exitproc();
    }
}
